import math
from collections.abc import Iterable
from functools import cached_property
from typing import Literal

import numpy as np
import numpy.typing as npt

from . import base
from .constellations import constellation_qam
from .labelings import get_labeling


class QAModulation(base.Modulation[np.complexfloating]):
    r"""
    Quadrature-amplitude modulation (QAM). It is a complex [modulation scheme](/ref/Modulation) in which the constellation is given as a Cartesian product of two [PAM](/ref/PAModulation) constellations, namely, the *in-phase constellation*, and the *quadrature constellation*. More precisely, the $i$-th constellation symbol is given by
    $$
        \begin{aligned}
            x_i = \left[ A_\mathrm{I} \left( 2i_\mathrm{I} - M_\mathrm{I} + 1 \right) + \mathrm{j} A_\mathrm{Q} \left( 2i_\mathrm{Q} - M_\mathrm{Q} + 1 \right) \right] \exp(\mathrm{j}\phi), \quad
                &  i \in [0 : M), \\\\
                & i_\mathrm{I} = i \bmod M_\mathrm{I}, \\\\
                & i_\mathrm{Q} = \lfloor i / M_\mathrm{I} \rfloor,
        \end{aligned}
    $$
    where $M_\mathrm{I}$ and $M_\mathrm{Q}$ are the *orders* (powers of $2$), and $A_\mathrm{I}$ and $A_\mathrm{Q}$ are the *base amplitudes* of the in-phase and quadrature constellations, respectively. Also, $\phi$ is the *phase offset*. The order of the resulting complex-valued constellation is $M = M_\mathrm{I} M_\mathrm{Q}$, a power of $2$.

    Parameters:
        orders: A tuple $(M_\mathrm{I}, M_\mathrm{Q})$ with the orders of the in-phase and quadrature constellations, respectively; both $M_\mathrm{I}$ and $M_\mathrm{Q}$ must be powers of $2$. If specified as a single integer $M$, then it is assumed that $M_\mathrm{I} = M_\mathrm{Q} = \sqrt{M}$; in this case, $M$ must be an square power of $2$.

        base_amplitudes: A tuple $(A_\mathrm{I}, A_\mathrm{Q})$ with the base amplitudes of the in-phase and quadrature constellations, respectively. If specified as a single float $A$, then it is assumed that $A_\mathrm{I} = A_\mathrm{Q} = A$. The default value is $1.0$.

        phase_offset: The phase offset $\phi$ of the constellation. The default value is `0.0`.

        labeling: The binary labeling of the modulation. Can be specified either as a 2D-array of integers (see [base class](/ref/Modulation) for details), or as a string. In the latter case, the string must be either `'natural_2d'` or `'reflected_2d'`. The default value is `'reflected_2d'`, corresponding to the Gray labeling.

    Examples:
        1. The square $16$-QAM modulation with $(M_\mathrm{I}, M_\mathrm{Q}) = (4, 4)$ and $(A_\mathrm{I}, A_\mathrm{Q}) = (1, 1)$, and Gray labeling is depicted below.
            <figure markdown>
            ![16-QAM modulation with Gray labeling.](/figures/qam_16_gray.svg)
            </figure>

                >>> qam = komm.QAModulation(16)
                >>> qam.constellation
                array([-3.-3.j, -1.-3.j,  1.-3.j,  3.-3.j,
                       -3.-1.j, -1.-1.j,  1.-1.j,  3.-1.j,
                       -3.+1.j, -1.+1.j,  1.+1.j,  3.+1.j,
                       -3.+3.j, -1.+3.j,  1.+3.j,  3.+3.j])
                >>> qam.labeling
                array([[0, 0, 0, 0], [0, 1, 0, 0], [1, 1, 0, 0], [1, 0, 0, 0],
                       [0, 0, 0, 1], [0, 1, 0, 1], [1, 1, 0, 1], [1, 0, 0, 1],
                       [0, 0, 1, 1], [0, 1, 1, 1], [1, 1, 1, 1], [1, 0, 1, 1],
                       [0, 0, 1, 0], [0, 1, 1, 0], [1, 1, 1, 0], [1, 0, 1, 0]])

        1.  The rectangular $8$-QAM modulation with $(M_\mathrm{I}, M_\mathrm{Q}) = (4, 2)$ and $(A_\mathrm{I}, A_\mathrm{Q}) = (1, 2)$, and natural labeling is depicted below.
            <figure markdown>
            ![8-QAM modulation with Gray labeling.](/figures/qam_8_natural.svg)
            </figure>

                >>> qam = komm.QAModulation(
                ...     orders=(4, 2),
                ...     base_amplitudes=(1.0, 2.0),
                ...     labeling="natural_2d"
                ... )
                >>> qam.constellation
                array([-3.-2.j, -1.-2.j,  1.-2.j,  3.-2.j,
                       -3.+2.j, -1.+2.j,  1.+2.j,  3.+2.j])
                >>> qam.labeling
                array([[0, 0, 0], [0, 1, 0], [1, 0, 0], [1, 1, 0],
                       [0, 0, 1], [0, 1, 1], [1, 0, 1], [1, 1, 1]])
    """

    def __init__(
        self,
        orders: tuple[int, int] | int,
        base_amplitudes: tuple[float, float] | float = 1.0,
        phase_offset: float = 0.0,
        labeling: (
            Literal["natural_2d", "reflected_2d"] | npt.ArrayLike
        ) = "reflected_2d",
    ) -> None:
        if isinstance(orders, Iterable):
            self._orders = orders
        else:
            if not math.isqrt(orders) ** 2 == orders:
                raise ValueError(
                    "when a single integer, 'orders' must be a square power of 2"
                )
            self._orders = (math.isqrt(orders), math.isqrt(orders))
        if isinstance(base_amplitudes, Iterable):
            self._base_amplitudes = base_amplitudes
        else:
            self._base_amplitudes = (base_amplitudes, base_amplitudes)
        self._phase_offset = phase_offset
        self._labeling = labeling
        super()._validate_parameters()

    def __repr__(self) -> str:
        if isinstance(self._labeling, str):
            labeling_repr = repr(self._labeling)
        else:
            labeling_repr = self.labeling.tolist()
        args = ", ".join([
            f"orders={self._orders}",
            f"base_amplitudes={self._base_amplitudes}",
            f"phase_offset={self._phase_offset}",
            f"labeling={labeling_repr}",
        ])
        return f"{self.__class__.__name__}({args})"

    @cached_property
    def constellation(self) -> npt.NDArray[np.complexfloating]:
        r"""
        Examples:
            >>> qam = komm.QAModulation(16)
            >>> qam.constellation
            array([-3.-3.j, -1.-3.j,  1.-3.j,  3.-3.j,
                   -3.-1.j, -1.-1.j,  1.-1.j,  3.-1.j,
                   -3.+1.j, -1.+1.j,  1.+1.j,  3.+1.j,
                   -3.+3.j, -1.+3.j,  1.+3.j,  3.+3.j])
        """
        return constellation_qam(
            self._orders, self._base_amplitudes, self._phase_offset
        )

    @cached_property
    def labeling(self) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> qam = komm.QAModulation(16)
            >>> qam.labeling
            array([[0, 0, 0, 0], [0, 1, 0, 0], [1, 1, 0, 0], [1, 0, 0, 0],
                   [0, 0, 0, 1], [0, 1, 0, 1], [1, 1, 0, 1], [1, 0, 0, 1],
                   [0, 0, 1, 1], [0, 1, 1, 1], [1, 1, 1, 1], [1, 0, 1, 1],
                   [0, 0, 1, 0], [0, 1, 1, 0], [1, 1, 1, 0], [1, 0, 1, 0]])
        """
        return get_labeling(
            self._labeling, ("natural_2d", "reflected_2d"), self._orders
        )

    @cached_property
    def inverse_labeling(self) -> dict[tuple[int, ...], int]:
        r"""
        Examples:
            >>> qam = komm.QAModulation(16)
            >>> qam.inverse_labeling
            {(0, 0, 0, 0): 0,  (0, 1, 0, 0): 1,  (1, 1, 0, 0): 2,  (1, 0, 0, 0): 3,
             (0, 0, 0, 1): 4,  (0, 1, 0, 1): 5,  (1, 1, 0, 1): 6,  (1, 0, 0, 1): 7,
             (0, 0, 1, 1): 8,  (0, 1, 1, 1): 9,  (1, 1, 1, 1): 10, (1, 0, 1, 1): 11,
             (0, 0, 1, 0): 12, (0, 1, 1, 0): 13, (1, 1, 1, 0): 14, (1, 0, 1, 0): 15}
        """
        return super().inverse_labeling

    @cached_property
    def order(self) -> int:
        r"""
        Examples:
            >>> qam = komm.QAModulation(16)
            >>> qam.order
            16
        """
        return self._orders[0] * self._orders[1]

    @cached_property
    def bits_per_symbol(self) -> int:
        r"""
        Examples:
            >>> qam = komm.QAModulation(16)
            >>> qam.bits_per_symbol
            4
        """
        return super().bits_per_symbol

    @cached_property
    def energy_per_symbol(self) -> float:
        r"""
        For the QAM, it is given by
        $$
            E_\mathrm{s} = \frac{A_\mathrm{I}^2}{3} \left( M_\mathrm{I}^2 - 1 \right) + \frac{A_\mathrm{Q}^2}{3} \left( M_\mathrm{Q}^2 - 1 \right).
        $$
        For the special case of a square QAM, it simplifies to
        $$
            E_\mathrm{s} = \frac{2A^2}{3}(M - 1).
        $$

        Examples:
            >>> qam = komm.QAModulation(16)
            >>> qam.energy_per_symbol
            10.0
        """
        A_I, A_Q = self._base_amplitudes
        M_I, M_Q = self._orders
        return (A_I**2) * (M_I**2 - 1) / 3 + (A_Q**2) * (M_Q**2 - 1) / 3

    @cached_property
    def energy_per_bit(self) -> float:
        r"""
        Examples:
            >>> qam = komm.QAModulation(16)
            >>> qam.energy_per_bit
            2.5
        """
        return super().energy_per_bit

    @cached_property
    def symbol_mean(self) -> complex:
        r"""
        For the QAM, it is given by
        $$
            \mu_\mathrm{s} = 0.
        $$

        Examples:
            >>> qam = komm.QAModulation(16)
            >>> qam.symbol_mean
            0j
        """
        return 0j

    @cached_property
    def minimum_distance(self) -> float:
        r"""
        For the QAM, it is given by
        $$
            d_\mathrm{min} = 2 \min(A_\mathrm{I}, A_\mathrm{Q}).
        $$
        For the special case of a square QAM, it simplifies to
        $$
            d_\mathrm{min} = 2 A.
        $$

        Examples:
            >>> qam = komm.QAModulation(16)
            >>> qam.minimum_distance
            2.0
        """
        return 2.0 * min(self._base_amplitudes)

    def modulate(self, input: npt.ArrayLike) -> npt.NDArray[np.complexfloating]:
        r"""
        Examples:
            >>> qam = komm.QAModulation(16)
            >>> qam.modulate([0, 0, 1, 1, 0, 0, 0, 1])
            array([-3.+1.j, -3.-1.j])
        """
        return super().modulate(input)

    def demodulate_hard(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        A_I, A_Q = self._base_amplitudes
        M_I, M_Q = self._orders
        input_I = np.real(np.multiply(input, np.exp(-1j * self._phase_offset))) / A_I
        input_Q = np.imag(np.multiply(input, np.exp(-1j * self._phase_offset))) / A_Q
        indices_I = np.clip(np.around((input_I + M_I - 1) / 2), 0, M_I - 1).astype(int)
        indices_Q = np.clip(np.around((input_Q + M_Q - 1) / 2), 0, M_Q - 1).astype(int)
        indices = indices_I + indices_Q * M_I
        hard_bits = np.reshape(self.labeling[indices], shape=-1)
        return hard_bits

    def demodulate_soft(
        self, input: npt.ArrayLike, snr: float = 1.0
    ) -> npt.NDArray[np.floating]:
        return super().demodulate_soft(input, snr)



from abc import ABC, abstractmethod
from functools import cache, cached_property
from itertools import combinations

import numpy as np
import numpy.typing as npt
from tqdm import tqdm

from .._util.bit_operations import bits_to_int, int_to_bits
from .._util.decorators import blockwise


class BlockCode(ABC):
    @cached_property
    @abstractmethod
    def length(self) -> int:
        r"""
        The length $n$ of the code.
        """
        raise NotImplementedError

    @cached_property
    @abstractmethod
    def dimension(self) -> int:
        r"""
        The dimension $k$ of the code.
        """
        raise NotImplementedError

    @cached_property
    @abstractmethod
    def redundancy(self) -> int:
        r"""
        The redundancy $m$ of the code.
        """
        raise NotImplementedError

    @cached_property
    @abstractmethod
    def rate(self) -> float:
        r"""
        The rate $R = k/n$ of the code.
        """
        return self.dimension / self.length

    @cached_property
    @abstractmethod
    def generator_matrix(self) -> npt.NDArray[np.integer]:
        r"""
        The generator matrix $G \in \mathbb{B}^{k \times n}$ of the code.
        """
        raise NotImplementedError

    @cached_property
    @abstractmethod
    def generator_matrix_right_inverse(self) -> npt.NDArray[np.integer]:
        raise NotImplementedError

    @cached_property
    @abstractmethod
    def check_matrix(self) -> npt.NDArray[np.integer]:
        r"""
        The check matrix $H \in \mathbb{B}^{m \times n}$ of the code.
        """
        raise NotImplementedError

    @abstractmethod
    def encode(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Applies the encoding mapping $\Enc : \mathbb{B}^k \to \mathbb{B}^n$ of the code. This method takes one or more sequences of messages and returns their corresponding codeword sequences.

        Parameters:
            input: The input sequence(s). Can be either a single sequence whose length is a multiple of $k$, or a multidimensional array where the last dimension is a multiple of $k$.

        Returns:
            output: The output sequence(s). Has the same shape as the input, with the last dimension expanded from $bk$ to $bn$, where $b$ is a positive integer.
        """

        @blockwise(self.dimension)
        def encode(u: npt.NDArray[np.integer]):
            v = u @ self.generator_matrix % 2
            return v

        return encode(input)

    @abstractmethod
    def project_word(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        @blockwise(self.length)
        def project(v: npt.NDArray[np.integer]):
            u = v @ self.generator_matrix_right_inverse % 2
            return u

        return project(input)

    @abstractmethod
    def inverse_encode(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Applies the inverse encoding partial mapping $\Enc^{-1} : \mathbb{B}^n \rightharpoonup \mathbb{B}^k$ of the code. This method takes one or more sequences of codewords and returns their corresponding message sequences.

        Parameters:
            input: The input sequence(s). Can be either a single sequence whose length is a multiple of $n$, or a multidimensional array where the last dimension is a multiple of $n$.

        Returns:
            output: The output sequence(s). Has the same shape as the input, with the last dimension contracted from $bn$ to $bk$, where $b$ is a positive integer.

        Raises:
            ValueError: If the input contains any invalid codewords.
        """
        s = self.check(input)
        if not np.all(s == 0):
            raise ValueError("one or more inputs in 'v' are not valid codewords")
        return self.project_word(input)

    @abstractmethod
    def check(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Applies the check mapping $\mathrm{Chk}: \mathbb{B}^n \to \mathbb{B}^m$ of the code. This method takes one or more sequences of received words and returns their corresponding syndrome sequences.

        Parameters:
            input: The input sequence(s). Can be either a single sequence whose length is a multiple of $n$, or a multidimensional array where the last dimension is a multiple of $n$.

        Returns:
            output: The output sequence(s). Has the same shape as the input, with the last dimension contracted from $bn$ to $bm$, where $b$ is a positive integer.
        """

        @blockwise(self.length)
        def check(r: npt.NDArray[np.integer]):
            s = r @ self.check_matrix.T % 2
            return s

        return check(input)

    @cache
    @abstractmethod
    def codewords(self) -> npt.NDArray[np.integer]:
        r"""
        Returns the codewords of the code. This is a $2^k \times n$ matrix whose rows are all the codewords. The codeword in row $i$ corresponds to the message obtained by expressing $i$ in binary with $k$ bits (MSB in the right).
        """
        batch_size = 1024
        k, n = self.dimension, self.length
        codewords = np.empty((2**k, n), dtype=int)
        for i in tqdm(
            range(0, 2**k, batch_size), desc="Generating codewords", delay=2.5
        ):
            batch_end = min(i + batch_size, 2**k)
            js = np.arange(i, batch_end, dtype=np.uint64).reshape(-1, 1).view(np.uint8)
            messages_batch = np.unpackbits(js, axis=1, count=k, bitorder="little")
            codewords[i:batch_end] = self.encode(messages_batch)
        return codewords

    @cache
    @abstractmethod
    def codeword_weight_distribution(self) -> npt.NDArray[np.integer]:
        r"""
        Returns the codeword weight distribution of the code. This is an array of shape $(n + 1)$ in which element in position $w$ is equal to the number of codewords of Hamming weight $w$, for $w \in [0 : n]$.
        """
        return np.bincount(np.sum(self.codewords(), axis=1), minlength=self.length + 1)

    @cache
    @abstractmethod
    def minimum_distance(self) -> int:
        r"""
        Returns the minimum distance $d$ of the code. This is equal to the minimum Hamming weight of the non-zero codewords.
        """
        return int(np.flatnonzero(self.codeword_weight_distribution())[1])

    @cache
    @abstractmethod
    def coset_leaders(self) -> npt.NDArray[np.integer]:
        r"""
        Returns the coset leaders of the code. This is a $2^m \times n$ matrix whose rows are all the coset leaders. The coset leader in row $i$ corresponds to the syndrome obtained by expressing $i$ in binary with $m$ bits (MSB in the right), and whose Hamming weight is minimal. This may be used as a LUT for syndrome-based decoding.
        """
        m, n = self.redundancy, self.length
        leaders = np.full(2**m, -1)
        taken = 0
        with tqdm(total=2**m, desc="Generating coset leaders", delay=2.5) as pbar:
            for w in range(n + 1):
                for idx in combinations(range(n), w):
                    leader = np.zeros(n, dtype=int)
                    leader[list(idx)] = 1
                    syndrome = self.check(leader)
                    i = bits_to_int(syndrome)
                    if leaders[i] != -1:
                        continue
                    taken += 1
                    pbar.update(1)
                    leaders[i] = bits_to_int(leader)
                    if taken == 2**m:
                        return int_to_bits(leaders, n)

    @cache
    @abstractmethod
    def coset_leader_weight_distribution(self) -> npt.NDArray[np.integer]:
        r"""
        Returns the coset leader weight distribution of the code. This is an array of shape $(n + 1)$ in which element in position $w$ is equal to the number of coset leaders of weight $w$, for $w \in [0 : n]$.
        """
        return np.bincount(
            np.sum(self.coset_leaders(), axis=1), minlength=self.length + 1
        )

    @cache
    @abstractmethod
    def packing_radius(self) -> int:
        r"""
        Returns the packing radius of the code. This is also called the *error-correcting capability* of the code, and is equal to $\lfloor (d - 1) / 2 \rfloor$.
        """
        return (self.minimum_distance() - 1) // 2

    @cache
    @abstractmethod
    def covering_radius(self) -> int:
        r"""
        Returns the covering radius of the code. This is equal to the maximum Hamming weight of the coset leaders.
        """
        return int(np.flatnonzero(self.coset_leader_weight_distribution())[-1])



from dataclasses import dataclass
from functools import cache
from heapq import heapify, heappop, heappush
from itertools import product
from math import prod
from typing import Literal

import numpy.typing as npt
from tqdm import tqdm
from typing_extensions import Self

from .._util.docs import mkdocstrings
from .._util.information_theory import PMF
from .FixedToVariableCode import FixedToVariableCode
from .util import Word, empty_mapping


@mkdocstrings(filters=["!.*"])
class HuffmanCode(FixedToVariableCode):
    r"""
    Binary Huffman code. It is an optimal (minimal expected rate) [fixed-to-variable length code](/ref/FixedToVariableCode) for a given probability mass function. For more details, see <cite>Say06, Sec. 3.2</cite>.

    Notes:
        Huffman codes are always [prefix-free](/ref/FixedToVariableCode/#is_prefix_free) (hence [uniquely decodable](/ref/FixedToVariableCode/#is_uniquely_decodable)).

    Parameters:
        pmf: The probability mass function of the source.
        source_block_size: The source block size $k$. The default value is $k = 1$.
        policy: The policy to be used when constructing the code. It must be either `'high'` (move combined symbols as high as possible) or `'low'` (move combined symbols as low as possible). The default value is `'high'`.

    Examples:
        >>> pmf = [0.7, 0.15, 0.15]

        >>> code = komm.HuffmanCode(pmf)
        >>> code.enc_mapping
        {(0,): (0,),
         (1,): (1, 1),
         (2,): (1, 0)}
        >>> code.rate(pmf)  # doctest: +FLOAT_CMP
        np.float64(1.3)

        >>> code = komm.HuffmanCode(pmf, 2)
        >>> code.enc_mapping
        {(0, 0): (1,),
         (0, 1): (0, 0, 0, 0),
         (0, 2): (0, 1, 1),
         (1, 0): (0, 1, 0),
         (1, 1): (0, 0, 0, 1, 1, 1),
         (1, 2): (0, 0, 0, 1, 1, 0),
         (2, 0): (0, 0, 1),
         (2, 1): (0, 0, 0, 1, 0, 1),
         (2, 2): (0, 0, 0, 1, 0, 0)}
        >>> code.rate(pmf)  # doctest: +FLOAT_CMP
        np.float64(1.1975)
    """

    def __init__(
        self,
        pmf: npt.ArrayLike,
        source_block_size: int = 1,
        policy: Literal["high", "low"] = "high",
    ):
        self.pmf = PMF(pmf)
        if not policy in {"high", "low"}:
            raise ValueError("'policy': must be in {'high', 'low'}")
        self.policy = policy
        super().__init__(
            source_cardinality=self.pmf.size,
            target_cardinality=2,
            source_block_size=source_block_size,
            enc_mapping=huffman_algorithm(self.pmf, source_block_size, policy),
        )

    def __repr__(self) -> str:
        args = ", ".join([
            f"pmf={self.pmf.tolist()}",
            f"source_block_size={self.source_block_size}",
            f"policy={self.policy!r}",
        ])
        return f"{self.__class__.__name__}({args})"

    @cache
    def is_uniquely_decodable(self) -> bool:
        return True

    @cache
    def is_prefix_free(self) -> bool:
        return True


def huffman_algorithm(
    pmf: PMF, source_block_size: int, policy: Literal["high", "low"]
) -> dict[Word, Word]:
    @dataclass
    class Node:
        index: int
        probability: float
        parent: int | None = None
        bit: int = -1

        def __lt__(self, other: Self) -> bool:
            i0, p0 = self.index, self.probability
            i1, p1 = other.index, other.probability
            if policy == "high":
                return (p0, i0) < (p1, i1)
            elif policy == "low":
                return (p0, -i0) < (p1, -i1)

    pbar = tqdm(
        desc="Generating Huffman code",
        total=3 * pmf.size**source_block_size - 1,
        delay=2.5,
    )

    tree: list[Node] = []
    for index, probs in enumerate(product(pmf, repeat=source_block_size)):
        tree.append(Node(index, prod(probs)))
        pbar.update()

    heap = tree.copy()
    heapify(heap)
    while len(heap) > 1:
        node1 = heappop(heap)
        node0 = heappop(heap)
        node1.bit = 1
        node0.bit = 0
        node = Node(index=len(tree), probability=node0.probability + node1.probability)
        node0.parent = node1.parent = node.index
        heappush(heap, node)
        tree.append(node)
        pbar.update()

    enc_mapping = empty_mapping(pmf.size, source_block_size)
    for index, u in enumerate(enc_mapping.keys()):
        node = tree[index]
        bits: list[int] = []
        while node.parent is not None:
            bits.append(node.bit)
            node = tree[node.parent]
        v = tuple(reversed(bits))
        enc_mapping[u] = v
        pbar.update()

    pbar.close()

    return enc_mapping



from dataclasses import dataclass

import numpy as np
import numpy.typing as npt

from .._util.bit_operations import int_to_bits
from . import base


@dataclass
class SlepianArray:
    r"""
    Slepian array (standard array) for a [linear block code](/ref/BlockCode). It is a table with $2^m$ rows and $2^k$ columns, where $m$ is the redundancy, and $k$ is the dimension of the code. Each row corresponds to a _coset_ of the group of codewords, in which:

    - The first row is the group of codewords itself.

    - The first column contains _coset leaders_ (i.e., elements of minimal weight in its coset).

    In this implementation:

    - A row's index $i$ corresponds to the $m$-bit syndrome obtained by expressing $i$ in binary (MSB on the right).

    - A column's index $j$ corresponds to the $k$-bit message obtained by expressing $j$ in binary (MSB on the right).

    For more details, see <cite>LC04, Sec. 3.5</cite>.

    Attributes:
        code: The linear block code for which the Slepian array is generated.
    """

    code: base.BlockCode

    def __post_init__(self) -> None:
        self._leaders = self.code.coset_leaders()

    def entry(self, i: int, j: int) -> npt.NDArray[np.integer]:
        r"""
        The entry at the $i$-th row and $j$-th column of the Slepian array.

        Parameters:
            i: The index of the row.
            j: The index of the column.

        Examples:
            >>> code = komm.BlockCode(generator_matrix=[[1, 0, 0, 0, 1, 1], [0, 1, 0, 1, 0, 1], [0, 0, 1, 1, 1, 0]])
            >>> sa = komm.SlepianArray(code)
            >>> binlist2str = lambda binlist: "".join(str(bit) for bit in binlist)
            >>> m, k = code.redundancy, code.dimension
            >>> for i in range(2**m):
            ...     for j in range(2**k):
            ...         print(binlist2str(sa.entry(i, j)), end=" ")
            ...     print()
            000000 100011 010101 110110 001110 101101 011011 111000
            000100 100111 010001 110010 001010 101001 011111 111100
            000010 100001 010111 110100 001100 101111 011001 111010
            001000 101011 011101 111110 000110 100101 010011 110000
            000001 100010 010100 110111 001111 101100 011010 111001
            010000 110011 000101 100110 011110 111101 001011 101000
            100000 000011 110101 010110 101110 001101 111011 011000
            100100 000111 110001 010010 101010 001001 111111 011100
        """
        k = self.code.dimension
        leader = self._leaders[i]
        message = int_to_bits(j, k)
        codeword = self.code.encode(message)
        return np.array(leader + codeword) % 2

    def row(self, i: int) -> npt.NDArray[np.integer]:
        r"""
        The $i$-th row of the Slepian array.

        Parameters:
            i: The index of the row.

        Examples:
            >>> code = komm.BlockCode(generator_matrix=[[1, 0, 0, 0, 1, 1], [0, 1, 0, 1, 0, 1], [0, 0, 1, 1, 1, 0]])
            >>> sa = komm.SlepianArray(code)
            >>> sa.row(0)  # The codewords
            array([[0, 0, 0, 0, 0, 0],
                   [1, 0, 0, 0, 1, 1],
                   [0, 1, 0, 1, 0, 1],
                   [1, 1, 0, 1, 1, 0],
                   [0, 0, 1, 1, 1, 0],
                   [1, 0, 1, 1, 0, 1],
                   [0, 1, 1, 0, 1, 1],
                   [1, 1, 1, 0, 0, 0]])
        """
        k = self.code.dimension
        return np.array([self.entry(i, j) for j in range(2**k)])

    def col(self, j: int) -> npt.NDArray[np.integer]:
        r"""
        The $j$-th column of the Slepian array.

        Parameters:
            j: The index of the column.

        Examples:
            >>> code = komm.BlockCode(generator_matrix=[[1, 0, 0, 0, 1, 1], [0, 1, 0, 1, 0, 1], [0, 0, 1, 1, 1, 0]])
            >>> sa = komm.SlepianArray(code)
            >>> sa.col(0)  # The coset leaders
            array([[0, 0, 0, 0, 0, 0],
                   [0, 0, 0, 1, 0, 0],
                   [0, 0, 0, 0, 1, 0],
                   [0, 0, 1, 0, 0, 0],
                   [0, 0, 0, 0, 0, 1],
                   [0, 1, 0, 0, 0, 0],
                   [1, 0, 0, 0, 0, 0],
                   [1, 0, 0, 1, 0, 0]])
        """
        m = self.code.redundancy
        return np.array([self.entry(i, j) for i in range(2**m)])



from functools import cached_property
from typing import Literal

import numpy as np
import numpy.typing as npt

from . import base
from .constellations import constellation_psk
from .labelings import get_labeling


class PSKModulation(base.Modulation[np.complexfloating]):
    r"""
    Phase-shift keying (PSK) modulation. It is a complex [modulation scheme](/ref/Modulation) in which the constellation symbols are *uniformly arranged* in a circle. More precisely, the the $i$-th constellation symbol is given by
    $$
        x_i = A \exp \left( \mathrm{j} \frac{2 \pi i}{M} \right) \exp(\mathrm{j} \phi), \quad i \in [0 : M),
    $$
    where $M$ is the *order* (a power of $2$), $A$ is the *amplitude*, and $\phi$ is the *phase offset* of the modulation.

    Parameters:
        order: The order $M$ of the modulation. It must be a power of $2$.

        amplitude: The amplitude $A$ of the constellation. The default value is `1.0`.

        phase_offset: The phase offset $\phi$ of the constellation. The default value is `0.0`.

        labeling: The binary labeling of the modulation. Can be specified either as a 2D-array of integers (see [base class](/ref/Modulation) for details), or as a string. In the latter case, the string must be either `'natural'` or `'reflected'`. The default value is `'reflected'`, corresponding to the Gray labeling.

    Examples:
        1. The $4$-PSK modulation with base amplitude $A = 1$, phase offset $\phi = 0$, and Gray labeling is depicted below.
            <figure markdown>
            ![4-PSK modulation with Gray labeling.](/figures/psk_4_gray.svg)
            </figure>

                >>> psk = komm.PSKModulation(4)
                >>> psk.constellation.round(3)
                array([ 1.+0.j,  0.+1.j, -1.+0.j, -0.-1.j])
                >>> psk.labeling
                array([[0, 0],
                       [0, 1],
                       [1, 1],
                       [1, 0]])

        1. The $8$-PSK modulation with base amplitude $A = 0.5$, phase offset $\phi = \pi/8$, and natural labeling is depicted below.
            <figure markdown>
            ![8-PSK modulation with natural labeling.](/figures/psk_8_natural.svg)
            </figure>

                >>> psk = komm.PSKModulation(
                ...     order=8,
                ...     amplitude=0.5,
                ...     phase_offset=np.pi/8,
                ...     labeling='natural',
                ... )
                >>> psk.constellation.round(3)
                array([ 0.462+0.191j,  0.191+0.462j, -0.191+0.462j, -0.462+0.191j,
                       -0.462-0.191j, -0.191-0.462j,  0.191-0.462j,  0.462-0.191j])
                >>> psk.labeling
                array([[0, 0, 0],
                       [0, 0, 1],
                       [0, 1, 0],
                       [0, 1, 1],
                       [1, 0, 0],
                       [1, 0, 1],
                       [1, 1, 0],
                       [1, 1, 1]])
    """

    def __init__(
        self,
        order: int,
        amplitude: float = 1.0,
        phase_offset: float = 0.0,
        labeling: Literal["natural", "reflected"] | npt.ArrayLike = "reflected",
    ) -> None:
        self._order = order
        self._amplitude = amplitude
        self._phase_offset = phase_offset
        self._labeling = labeling
        super()._validate_parameters()

    def __repr__(self) -> str:
        if isinstance(self._labeling, str):
            labeling_repr = repr(self._labeling)
        else:
            labeling_repr = self.labeling.tolist()
        args = ", ".join([
            f"order={self._order}",
            f"amplitude={self._amplitude}",
            f"phase_offset={self._phase_offset}",
            f"labeling={labeling_repr}",
        ])
        return f"{self.__class__.__name__}({args})"

    @cached_property
    def constellation(self) -> npt.NDArray[np.complexfloating]:
        r"""
        Examples:
            >>> psk = komm.PSKModulation(4)
            >>> psk.constellation.round(3)
            array([ 1.+0.j,  0.+1.j, -1.+0.j, -0.-1.j])
        """
        return constellation_psk(self._order, self._amplitude, self._phase_offset)

    @cached_property
    def labeling(self) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> psk = komm.PSKModulation(4)
            >>> psk.labeling
            array([[0, 0],
                   [0, 1],
                   [1, 1],
                   [1, 0]])
        """
        return get_labeling(self._labeling, ("natural", "reflected"), self._order)

    @cached_property
    def inverse_labeling(self) -> dict[tuple[int, ...], int]:
        r"""
        Examples:
            >>> psk = komm.PSKModulation(4)
            >>> psk.inverse_labeling
            {(0, 0): 0, (0, 1): 1, (1, 1): 2, (1, 0): 3}
        """
        return super().inverse_labeling

    @cached_property
    def order(self) -> int:
        r"""
        Examples:
            >>> psk = komm.PSKModulation(4)
            >>> psk.order
            4
        """
        return self._order

    @cached_property
    def bits_per_symbol(self) -> int:
        r"""
        Examples:
            >>> psk = komm.PSKModulation(4)
            >>> psk.bits_per_symbol
            2
        """
        return super().bits_per_symbol

    @cached_property
    def energy_per_symbol(self) -> float:
        r"""
        For the PSK, it is given by
        $$
            E_\mathrm{s} = A^2.
        $$

        Examples:
            >>> psk = komm.PSKModulation(4)
            >>> psk.energy_per_symbol
            1.0
        """
        return self._amplitude**2

    @cached_property
    def energy_per_bit(self) -> float:
        r"""
        Examples:
            >>> psk = komm.PSKModulation(4)
            >>> psk.energy_per_bit
            0.5
        """
        return super().energy_per_bit

    @cached_property
    def symbol_mean(self) -> complex:
        r"""
        For the PSK, it is given by
        $$
            \mu_\mathrm{s} = 0.
        $$

        Examples:
            >>> psk = komm.PSKModulation(4)
            >>> psk.symbol_mean
            0j
        """
        return 0j

    @cached_property
    def minimum_distance(self) -> float:
        r"""
        For the PSK, it is given by
        $$
            d_\mathrm{min} = 2A\sin\left(\frac{\pi}{M}\right).
        $$

        Examples:
            >>> psk = komm.PSKModulation(4)
            >>> psk.minimum_distance  # doctest: +FLOAT_CMP
            1.414213562373095
        """
        return 2.0 * self._amplitude * float(np.sin(np.pi / self._order))

    def modulate(self, input: npt.ArrayLike) -> npt.NDArray[np.complexfloating]:
        r"""
        Examples:
            >>> psk = komm.PSKModulation(4)
            >>> psk.modulate([0, 0, 1, 1, 0, 0, 0, 1]).round(3)  # doctest: +FLOAT_CMP
            array([ 1.+0.j, -1.+0.j,  1.+0.j,  0.+1.j])
        """
        return super().modulate(input)

    def demodulate_hard(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        M, phi = self._order, self._phase_offset
        turn = np.angle(np.multiply(input, np.exp(-1j * phi))) / (2 * np.pi)
        indices = np.mod(np.around(M * turn), M).astype(int)
        hard_bits = np.reshape(self.labeling[indices], shape=-1)
        return hard_bits

    def demodulate_soft(
        self, input: npt.ArrayLike, snr: float = 1.0
    ) -> npt.NDArray[np.floating]:
        return super().demodulate_soft(input, snr)



from dataclasses import dataclass
from functools import cached_property

import numpy as np
import numpy.typing as npt

from . import base
from .RaisedCosinePulse import RaisedCosinePulse


@dataclass
class RootRaisedCosinePulse(base.Pulse):
    r"""
    Root raised cosine pulse. It is a [pulse](/ref/Pulse) with waveform given by
    $$
        h(t) = \frac{\sin \( 2 \pi f_1 t \) + 4 \alpha t \cos \( 2 \pi f_2 t \)}{\pi t \( 1 - (4 \alpha t)^2 \)},
    $$
    where $\alpha$ is the *roll-off factor* (which must satisfy $0 \leq \alpha \leq 1$), $f_1 = (1 - \alpha) / 2$, and $f_2 = (1 + \alpha) / 2$. Its spectrum is given by the square root of the spectrum of the [raised cosine pulse](/ref/RaisedCosinePulse).

    The root raised cosine pulse is depicted below for $\alpha = 0.25$, and for $\alpha = 0.75$.

    <div class="centered" markdown>
      <span>
        ![Root raised cosine pulse with roll-off factor 0.25.](/figures/pulse_root_raised_cosine_25.svg)
      </span>
      <span>
        ![Root raised cosine pulse with roll-off factor 0.75.](/figures/pulse_root_raised_cosine_75.svg)
      </span>
    </div>

    Attributes:
        rolloff: The roll-off factor $\alpha$ of the pulse. Must satisfy $0 \leq \alpha \leq 1$. The default value is `0.0`.

    Examples:
        >>> pulse = komm.RootRaisedCosinePulse(rolloff=0.25)
        >>> pulse.waveform([-0.75, -0.50, -0.25,  0.00,  0.25,  0.50,  0.75]).round(4)
        array([0.2379, 0.6218, 0.9432, 1.0683, 0.9432, 0.6218, 0.2379])
        >>> pulse.spectrum([-0.75, -0.50, -0.25,  0.00,  0.25,  0.50,  0.75]).round(4)
        array([0.    , 0.7071, 1.    , 1.    , 1.    , 0.7071, 0.    ])
    """

    rolloff: float = 0.0

    def waveform(self, t: npt.ArrayLike) -> npt.NDArray[np.floating]:
        a = self.rolloff
        t = np.asarray(t) + 1e-8  # TODO: Improve this workaround
        f1 = (1 - a) / 2
        f2 = (1 + a) / 2
        num = np.sin(2 * np.pi * f1 * t) + 4 * a * t * np.cos(2 * np.pi * f2 * t)
        den = np.pi * t * (1 - (4 * a * t) ** 2)
        return num / den

    def spectrum(self, f: npt.ArrayLike) -> npt.NDArray[np.floating]:
        a = self.rolloff
        hf_rc = RaisedCosinePulse(rolloff=a).spectrum
        return np.sqrt(hf_rc(f))

    @cached_property
    def support(self) -> tuple[float, float]:
        return (-np.inf, np.inf)



from collections.abc import Callable
from dataclasses import dataclass
from functools import cached_property

import numpy as np
import numpy.typing as npt

from . import base


@dataclass
class LloydMaxQuantizer(base.ScalarQuantizer):
    r"""
    Lloyd–Max scalar quantizer. It is a [scalar quantizer](/ref/ScalarQuantizer) that minimizes the mean-squared error (MSE) between the input signal $X$ and its quantized version. For more details, see <cite>Say06, Sec. 9.6.1</cite>.

    Parameters:
        input_pdf: The probability density function $f_X(x)$ of the input signal.

        input_range: The range $(x_\mathrm{min}, x_\mathrm{max})$ of the input signal.

        num_levels: The number $L$ of quantization levels. It must be greater than $1$.

    Examples:
        >>> uniform_pdf = lambda x: 1/8 * (np.abs(x) <= 4)
        >>> quantizer = komm.LloydMaxQuantizer(
        ...     input_pdf=uniform_pdf,
        ...     input_range=(-4, 4),
        ...     num_levels=8,
        ... )
        >>> quantizer.levels
        array([-3.5, -2.5, -1.5, -0.5,  0.5,  1.5,  2.5,  3.5])
        >>> quantizer.thresholds
        array([-3., -2., -1.,  0.,  1.,  2.,  3.])

        >>> gaussian_pdf = lambda x: 1/np.sqrt(2*np.pi) * np.exp(-x**2/2)
        >>> quantizer = komm.LloydMaxQuantizer(
        ...     input_pdf=gaussian_pdf,
        ...     input_range=(-5, 5),
        ...     num_levels=8,
        ... )
        >>> quantizer.levels.round(3)
        array([-2.152, -1.344, -0.756, -0.245,  0.245,  0.756,  1.344,  2.152])
        >>> quantizer.thresholds.round(3)  # doctest: +FLOAT_CMP
        array([-1.748, -1.05 , -0.501,  0.   ,  0.501,  1.05 ,  1.748])
    """

    input_pdf: Callable[[npt.NDArray[np.floating]], npt.NDArray[np.floating]]
    input_range: tuple[float, float]
    num_levels: int

    def __post_init__(self) -> None:
        self._levels, self._thresholds = lloyd_max_quantizer(
            self.input_pdf,
            self.num_levels,
            self.input_range,
            points_per_interval=1000,
            max_iter=100,
        )

    @cached_property
    def levels(self) -> npt.NDArray[np.floating]:
        r"""
        Examples:
            >>> gaussian_pdf = lambda x: 1/np.sqrt(2*np.pi) * np.exp(-x**2/2)
            >>> quantizer = komm.LloydMaxQuantizer(
            ...     input_pdf=gaussian_pdf,
            ...     input_range=(-5, 5),
            ...     num_levels=8,
            ... )
            >>> quantizer.levels.round(3)
            array([-2.152, -1.344, -0.756, -0.245,  0.245,  0.756,  1.344,  2.152])
        """
        return self._levels

    @cached_property
    def thresholds(self) -> npt.NDArray[np.floating]:
        r"""
        Examples:
            >>> gaussian_pdf = lambda x: 1/np.sqrt(2*np.pi) * np.exp(-x**2/2)
            >>> quantizer = komm.LloydMaxQuantizer(
            ...     input_pdf=gaussian_pdf,
            ...     input_range=(-5, 5),
            ...     num_levels=8,
            ... )
            >>> quantizer.thresholds.round(3)  # doctest: +FLOAT_CMP
            array([-1.748, -1.05 , -0.501,  0.   ,  0.501,  1.05 ,  1.748])
        """
        return self._thresholds

    def digitize(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> gaussian_pdf = lambda x: 1/np.sqrt(2*np.pi) * np.exp(-x**2/2)
            >>> quantizer = komm.LloydMaxQuantizer(
            ...     input_pdf=gaussian_pdf,
            ...     input_range=(-5, 5),
            ...     num_levels=8,
            ... )
            >>> quantizer.digitize([0, 1, 2, 3, 4, 5, 6, 7])
            array([4, 5, 7, 7, 7, 7, 7, 7])
        """
        return super().digitize(input)

    def quantize(self, input: npt.ArrayLike) -> npt.NDArray[np.floating]:
        r"""
        Examples:
            >>> gaussian_pdf = lambda x: 1/np.sqrt(2*np.pi) * np.exp(-x**2/2)
            >>> quantizer = komm.LloydMaxQuantizer(
            ...     input_pdf=gaussian_pdf,
            ...     input_range=(-5, 5),
            ...     num_levels=8,
            ... )
            >>> quantizer.quantize([0, 1, 2, 3, 4, 5, 6, 7]).round(3)
            array([0.245, 0.756, 2.152, 2.152, 2.152, 2.152, 2.152, 2.152])
        """
        return super().quantize(input)


def lloyd_max_quantizer(
    input_pdf: Callable[[npt.NDArray[np.floating]], npt.NDArray[np.floating]],
    num_levels: int,
    input_range: tuple[float, float],
    points_per_interval: int,
    max_iter: int,
) -> tuple[npt.NDArray[np.floating], npt.NDArray[np.floating]]:
    # See [Say06, eqs. (9.27) and (9.28)].
    x_min, x_max = input_range
    delta = (x_max - x_min) / num_levels

    # Initial guess
    levels = np.linspace(x_min + delta / 2, x_max - delta / 2, num=num_levels)
    thresholds = np.empty(num_levels + 1, dtype=float)
    new_levels = np.empty_like(levels)

    for _ in range(max_iter):
        thresholds[0] = x_min
        thresholds[1:-1] = 0.5 * (levels[:-1] + levels[1:])
        thresholds[-1] = x_max

        for i in range(num_levels):
            left, right = thresholds[i], thresholds[i + 1]
            x = np.linspace(left, right, num=points_per_interval, dtype=float)
            pdf = input_pdf(x)
            numerator = np.trapezoid(x * pdf, x)
            denominator = np.trapezoid(pdf, x)
            if denominator != 0:
                new_levels[i] = numerator / denominator
            else:  # Keep old level
                new_levels[i] = levels[i]
        if np.allclose(levels, new_levels):
            break
        levels = new_levels.copy()

    return new_levels, thresholds[1:-1]



from dataclasses import dataclass

import numpy as np
import numpy.typing as npt

from .._error_control_convolutional import TerminatedConvolutionalCode
from .._util.bit_operations import int_to_bits
from .._util.decorators import blockwise, vectorize
from . import base


@dataclass
class BCJRDecoder(base.BlockDecoder[TerminatedConvolutionalCode]):
    r"""
    Bahl–Cocke–Jelinek–Raviv (BCJR) decoder for [terminated convolutional codes](/ref/TerminatedConvolutionalCode). For more details, see <cite>LC04, Sec. 12.6</cite>.

    Parameters:
        code: The terminated convolutional code to be used for decoding.
        snr: The signal-to-noise ratio (SNR) of the channel (linear, not decibel).

    Notes:
        - Input type: `soft`.
        - Output type: `soft`.
    """

    code: TerminatedConvolutionalCode
    snr: float = 1.0

    def __post_init__(self) -> None:
        self._fsm = self.code.convolutional_code.finite_state_machine()
        num_states = self._fsm.num_states
        if self.code.mode == "direct-truncation":
            self._initial_state_distribution = np.eye(1, num_states, 0)
            self._final_state_distribution = np.ones(num_states) / num_states
        elif self.code.mode == "zero-termination":
            self._initial_state_distribution = np.eye(1, num_states, 0)
            self._final_state_distribution = np.eye(1, num_states, 0)
        elif self.code.mode == "tail-biting":
            raise NotImplementedError("algorithm not implemented for 'tail-biting'")

        n = self.code.convolutional_code.num_output_bits
        self._cache_polar = (-1) ** int_to_bits(range(2**n), width=n)

    def _metric_function(self, y: int, z: float) -> float:
        return 2.0 * self.snr * np.dot(self._cache_polar[y], z)

    def __call__(self, input: npt.ArrayLike) -> npt.NDArray[np.integer | np.floating]:
        r"""
        Examples:
            >>> convolutional_code = komm.ConvolutionalCode(feedforward_polynomials=[[0b11, 0b1]], feedback_polynomials=[0b11, 0b11])
            >>> code = komm.TerminatedConvolutionalCode(convolutional_code, num_blocks=3, mode="zero-termination")
            >>> decoder = komm.BCJRDecoder(code, snr=0.25)
            >>> decoder([-0.8, -0.1, -1.0, +0.5, +1.8, -1.1, -1.6, +1.6])
            array([-0.47774884, -0.61545527,  1.03018771])
        """
        n = self.code.convolutional_code.num_output_bits
        mu = self.code.convolutional_code.memory_order

        @blockwise(self.code.length)
        @vectorize
        def decode(r: npt.NDArray[np.floating]):
            input_posteriors = self._fsm.forward_backward(
                observed_sequence=r.reshape(-1, n),
                metric_function=self._metric_function,
                initial_state_distribution=self._initial_state_distribution,
                final_state_distribution=self._final_state_distribution,
            )
            if self.code.mode == "zero-termination":
                input_posteriors = input_posteriors[:-mu]
            return np.log(input_posteriors[:, 0] / input_posteriors[:, 1])

        return decode(input)
